@@ -17,12 +17,14 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
 
   USER_AGENT = "PMT/1.1.1 (v3; Net::HTTP)".freeze
 
+  # From https://forgeapi.puppet.com/#!/release/getReleases
+  MODULE_RELEASE_EXCLUSIONS=%w[readme changelog license uri module tags supported file_size downloads created_at updated_at deleted_at].join(',').freeze
+
   attr_reader :host, :repository
 
-  def initialize(host = Puppet[:module_repository], strict_semver = true)
+  def initialize(host = Puppet[:module_repository])
     @host = host
     @repository = Puppet::Forge::Repository.new(host, USER_AGENT)
-    @strict_semver = strict_semver
   end
 
   # Return a list of module metadata hashes that match the search query.
@@ -56,14 +58,14 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
     matches = []
     uri = "/v3/modules?query=#{term}"
     if Puppet[:module_groups]
-      uri += "&module_groups=#{Puppet[:module_groups].gsub('+', ' ')}"
+      uri += "&module_groups=#{Puppet[:module_groups].tr('+', ' ')}"
     end
 
     while uri
       # make_http_request URI encodes parameters
       response = make_http_request(uri)
 
-      if response.code == '200'
+      if response.code == 200
         result = Puppet::Util::Json.load(response.body)
         uri = decode_uri(result['pagination']['next'])
         matches.concat result['results']
@@ -90,9 +92,9 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
   # @see SemanticPuppet::Dependency::Source#fetch
   def fetch(input)
     name = input.tr('/', '-')
-    uri = "/v3/releases?module=#{name}&sort_by=version"
+    uri = "/v3/releases?module=#{name}&sort_by=version&exclude_fields=#{MODULE_RELEASE_EXCLUSIONS}"
     if Puppet[:module_groups]
-      uri += "&module_groups=#{Puppet[:module_groups].gsub('+', ' ')}"
+      uri += "&module_groups=#{Puppet[:module_groups].tr('+', ' ')}"
     end
     releases = []
 
@@ -100,7 +102,7 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
       # make_http_request URI encodes parameters
       response = make_http_request(uri)
 
-      if response.code == '200'
+      if response.code == 200
         response = Puppet::Util::Json.load(response.body)
       else
         raise ResponseError.new(:uri => URI.parse(@host).merge(uri), :response => response)
@@ -120,7 +122,7 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
   class ModuleRelease < SemanticPuppet::Dependency::ModuleRelease
     attr_reader :install_dir, :metadata
 
-    def initialize(source, data, strict_semver = true)
+    def initialize(source, data)
       @data = data
       @metadata = meta = data['metadata']
 
@@ -132,7 +134,7 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
         dependencies = meta['dependencies'].collect do |dep|
           begin
             Puppet::ModuleTool::Metadata.new.add_dependency(dep['name'], dep['version_requirement'], dep['repository'])
-            Puppet::ModuleTool.parse_module_dependency(release, dep, strict_semver)[0..1]
+            Puppet::ModuleTool.parse_module_dependency(release, dep)[0..1]
           rescue ArgumentError => e
             raise ArgumentError, _("Malformed dependency: %{name}.") % { name: dep['name'] } +
                 ' ' + _("Exception was: %{detail}") % { detail: e }
@@ -170,7 +172,18 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
       Puppet.warning "#{@metadata['name']} has been deprecated by its author! View module on Puppet Forge for more info." if deprecated?
 
       download(@data['file_uri'], tmpfile)
-      validate_checksum(tmpfile, @data['file_md5'])
+      checksum = @data['file_sha256']
+      if checksum
+        validate_checksum(tmpfile, checksum, Digest::SHA256)
+      else
+        checksum = @data['file_md5']
+        if checksum
+          validate_checksum(tmpfile, checksum, Digest::MD5)
+        else
+          raise _("Forge module is missing SHA256 and MD5 checksums")
+        end
+      end
+
       unpack(tmpfile, tmpdir)
 
       @unpacked_into = Pathname.new(tmpdir)
@@ -194,14 +207,18 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
     def download(uri, destination)
       response = @source.make_http_request(uri, destination)
       destination.flush and destination.close
-      unless response.code == '200'
+      unless response.code == 200
         raise Puppet::Forge::Errors::ResponseError.new(:uri => uri, :response => response)
       end
     end
 
-    def validate_checksum(file, checksum)
-      if Digest::MD5.file(file.path).hexdigest != checksum
-        raise RuntimeError, _("Downloaded release for %{name} did not match expected checksum") % { name: name }
+    def validate_checksum(file, checksum, digest_class)
+      if Facter.value(:fips_enabled) && digest_class == Digest::MD5
+        raise _("Module install using MD5 is prohibited in FIPS mode.")
+      end
+
+      if digest_class.file(file.path).hexdigest != checksum
+        raise RuntimeError, _("Downloaded release for %{name} did not match expected checksum %{checksum}") % { name: name, checksum: checksum }
       end
     end
 
@@ -224,7 +241,7 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
     l = list.map do |release|
       metadata = release['metadata']
       begin
-        ModuleRelease.new(self, release, @strict_semver)
+        ModuleRelease.new(self, release)
       rescue ArgumentError => e
         Puppet.warning _("Cannot consider release %{name}-%{version}: %{error}") % { name: metadata['name'], version: metadata['version'], error: e }
         false
@@ -237,6 +254,6 @@ class Puppet::Forge < SemanticPuppet::Dependency::Source
   def decode_uri(uri)
     return if uri.nil?
 
-    URI.decode(uri.gsub('+', ' '))
+    Puppet::Util.uri_unescape(uri.tr('+', ' '))
   end
 end
