@@ -1,6 +1,7 @@
 require 'puppet/util'
 require 'monitor'
 require 'puppet/parser/parser_factory'
+require 'puppet/concurrent/lock'
 
 # Just define it, so this class has fewer load dependencies.
 class Puppet::Node
@@ -25,11 +26,6 @@ class Puppet::Node::Environment
   NO_MANIFEST = :no_manifest
 
   # The create() factory method should be used instead.
-  #
-  # @api private
-  def self.new(*args)
-    create(*args)
-  end
   private_class_method :new
 
   # Create a new environment with the given name
@@ -44,13 +40,7 @@ class Puppet::Node::Environment
   #
   # @api public
   def self.create(name, modulepath, manifest = NO_MANIFEST, config_version = nil)
-    obj = self.allocate
-    obj.send(:initialize,
-             name.intern,
-             expand_dirs(extralibs() + modulepath),
-             manifest == NO_MANIFEST ? manifest : Puppet::FileSystem.expand_path(manifest),
-             config_version)
-    obj
+    new(name, modulepath, manifest, config_version)
   end
 
   # A remote subclass to make it easier to trace instances when debugging.
@@ -81,11 +71,12 @@ class Puppet::Node::Environment
   #
   # @param name [Symbol] The environment name
   def initialize(name, modulepath, manifest, config_version)
-    @name = name
-    @modulepath = modulepath
-    @manifest = manifest
+    @lock = Puppet::Concurrent::Lock.new
+    @name = name.intern
+    @modulepath = self.class.expand_dirs(self.class.extralibs() + modulepath)
+    @manifest = manifest == NO_MANIFEST ? manifest : Puppet::FileSystem.expand_path(manifest)
+
     @config_version = config_version
-    @modules_strict_semver = false
   end
 
   # Creates a new Puppet::Node::Environment instance, overriding any of the passed
@@ -171,6 +162,10 @@ class Puppet::Node::Environment
   # @api private
   attr_accessor :loaders
 
+  # Lock for compilation that needs exclusive access to the environment
+  # @api private
+  attr_reader :lock
+
   # Checks to make sure that this environment did not have a manifest set in
   # its original environment.conf if Puppet is configured with
   # +disable_per_environment_manifest+ set true.  If it did, the environment's
@@ -240,30 +235,16 @@ class Puppet::Node::Environment
     Puppet.settings.value(param, self.name)
   end
 
-  # A SemanticPuppet::VersionRange version >= 1.0.0 will not include versions with pre-release
-  # identifiers unless that is explicitly declared. This may cause backward compatibility
-  # issues when resolving module dependencies and the flag is therefore set to `false` by default.
-  #
-  # @param flag [Boolean] set to true to resolve module dependencies using strict SemVer semantics
-  #
-  def modules_strict_semver=(flag)
-    @modules_strict_semver = flag
-  end
-
-  # @return [Boolean] the current value of the modules_strict_semver flag.
-  # @api public
-  def modules_strict_semver?
-    @modules_strict_semver
-  end
-
   # @api public
   # @return [Puppet::Resource::TypeCollection] The current global TypeCollection
   def known_resource_types
-    if @known_resource_types.nil?
-      @known_resource_types = Puppet::Resource::TypeCollection.new(self)
-      @known_resource_types.import_ast(perform_initial_import(), '')
+    @lock.synchronize do
+      if @known_resource_types.nil?
+        @known_resource_types = Puppet::Resource::TypeCollection.new(self)
+        @known_resource_types.import_ast(perform_initial_import(), '')
+      end
+      @known_resource_types
     end
-    @known_resource_types
   end
 
   # Yields each modules' plugin directory if the plugin directory (modulename/lib)
@@ -316,7 +297,13 @@ class Puppet::Node::Environment
   def modules
     if @modules.nil?
       module_references = []
-      seen_modules = {}
+      project = Puppet.lookup(:bolt_project) { nil }
+      seen_modules = if project && project.load_as_module?
+                       module_references << project.to_h
+                       { project.name => true }
+                     else
+                       {}
+                     end
       modulepath.each do |path|
         Dir.entries(path).each do |name|
           next unless Puppet::Module.is_module_directory?(name, path)
@@ -330,7 +317,7 @@ class Puppet::Node::Environment
 
       @modules = module_references.collect do |reference|
         begin
-          Puppet::Module.new(reference[:name], reference[:path], self, modules_strict_semver?)
+          Puppet::Module.new(reference[:name], reference[:path], self)
         rescue Puppet::Module::Error => e
           Puppet.log_exception(e)
           nil
@@ -354,9 +341,10 @@ class Puppet::Node::Environment
   # @param name [String] The directory name
   def warn_about_mistaken_path(path, name)
     if name == "lib"
-      Puppet.debug("Warning: Found directory named 'lib' in module path ('#{path}/lib'); unless " +
-          "you are expecting to load a module named 'lib', your module path may be set " +
-          "incorrectly.")
+      Puppet.debug {
+        "Warning: Found directory named 'lib' in module path ('#{path}/lib'); unless you \
+        are expecting to load a module named 'lib', your module path may be set incorrectly."
+      }
     end
   end
 
@@ -378,7 +366,7 @@ class Puppet::Node::Environment
             Puppet::Module.is_module_directory?(name, path)
           end
           modules_by_path[path] = module_names.sort.map do |name|
-            Puppet::Module.new(name, File.join(path, name), self, modules_strict_semver?)
+            Puppet::Module.new(name, File.join(path, name), self)
           end
         end
       else
@@ -462,9 +450,11 @@ class Puppet::Node::Environment
   # Checks if a reparse is required (cache of files is stale).
   #
   def check_for_reparse
-    if (Puppet[:code] != @parsed_code || @known_resource_types.parse_failed?)
-      @parsed_code = nil
-      @known_resource_types = nil
+    @lock.synchronize do
+      if (Puppet[:code] != @parsed_code || @known_resource_types.parse_failed?)
+        @parsed_code = nil
+        @known_resource_types = nil
+      end
     end
   end
 
@@ -522,14 +512,14 @@ class Puppet::Node::Environment
     end
   end
 
-  private
-
+  # not private so it can be called in initialize
   def self.expand_dirs(dirs)
     dirs.collect do |dir|
       Puppet::FileSystem.expand_path(dir)
     end
   end
-  private_class_method :expand_dirs
+
+  private
 
   # Reparse the manifests for the given environment
   #
