@@ -1,3 +1,5 @@
+require 'puppet/concurrent/thread_local_singleton'
+
 module Puppet::Pops
 module Types
 # The TypeCalculator can answer questions about puppet types.
@@ -99,6 +101,7 @@ module Types
 # @api public
 #
 class TypeCalculator
+  extend Puppet::Concurrent::ThreadLocalSingleton
 
   # @api public
   def self.assignable?(t1, t2)
@@ -108,29 +111,42 @@ class TypeCalculator
   # Answers, does the given callable accept the arguments given in args (an array or a tuple)
   # @param callable [PCallableType] - the callable
   # @param args [PArrayType, PTupleType] args optionally including a lambda callable at the end
-  # @return [Boolan] true if the callable accepts the arguments
+  # @return [Boolean] true if the callable accepts the arguments
   #
   # @api public
   def self.callable?(callable, args)
     singleton.callable?(callable, args)
   end
 
-  # Produces a String representation of the given type.
-  # @param t [PAnyType] the type to produce a string form
-  # @return [String] the type in string form
-  #
-  # @api public
-  #
-  def self.string(t)
-    Puppet.deprecation_warning('TypeCalculator.string is deprecated. Use TypeFormatter')
-    TypeFormatter.string(t)
-  end
-
   # @api public
   def self.infer(o)
     singleton.infer(o)
   end
 
+  # Infers a type if given object may have callable members, else returns nil.
+  # Caller must check for nil or if returned type supports members.
+  # This is a much cheaper call than doing a call to the general infer(o) method.
+  #
+  # @api private
+  def self.infer_callable_methods_t(o)
+    # If being a value that cannot have Pcore based methods callable from Puppet Language
+    if (o.is_a?(String) ||
+      o.is_a?(Numeric) ||
+      o.is_a?(TrueClass) ||
+      o.is_a?(FalseClass) ||
+      o.is_a?(Regexp) ||
+      o.instance_of?(Array) ||
+      o.instance_of?(Hash) ||
+      Types::PUndefType::DEFAULT.instance?(o)
+      )
+      return nil
+    end
+    # For other objects (e.g. PObjectType instances, and runtime types) full inference needed, since that will
+    # cover looking into the runtime type registry.
+    #
+    infer(o)
+  end
+
   # @api public
   def self.generalize(o)
     singleton.generalize(o)
@@ -141,58 +157,16 @@ class TypeCalculator
     singleton.infer_set(o)
   end
 
-  # @api public
-  def self.debug_string(t)
-    Puppet.deprecation_warning('TypeCalculator.debug_string is deprecated. Use TypeFormatter')
-    TypeFormatter.debug_string(t)
-  end
-
-  # @api public
-  def self.enumerable(t)
-    Puppet.deprecation_warning('TypeCalculator.enumerable is deprecated. Use iterable')
-    singleton.iterable(t)
-  end
-
   # @api public
   def self.iterable(t)
     singleton.iterable(t)
   end
 
-  # @return [TypeCalculator] the singleton instance
-  #
-  # @api private
-  def self.singleton
-    @tc_instance ||= new
-  end
-
   # @api public
   #
   def initialize
-    @@infer_visitor ||= Visitor.new(nil, 'infer',0,0)
-    @@extract_visitor ||= Visitor.new(nil, 'extract',0,0)
-  end
-
-  # Answers the question 'is it possible to inject an instance of the given class'
-  # A class is injectable if it has a special *assisted inject* class method called `inject` taking
-  # an injector and a scope as argument, or if it has a zero args `initialize` method.
-  #
-  # @param klazz [Class, PRuntimeType] the class/type to check if it is injectable
-  # @return [Class, nil] the injectable Class, or nil if not injectable
-  # @api public
-  #
-  def injectable_class(klazz)
-    # Handle case when we get a PType instead of a class
-    if klazz.is_a?(PRuntimeType)
-      klazz = ClassLoader.provide(klazz)
-    end
-
-    # data types can not be injected (check again, it is not safe to assume that given RubyRuntime klazz arg was ok)
-    return false unless type(klazz).is_a?(PRuntimeType)
-    if (klazz.respond_to?(:inject) && klazz.method(:inject).arity == -4) || klazz.instance_method(:initialize).arity == 0
-      klazz
-    else
-      nil
-    end
+    @infer_visitor = Visitor.new(nil, 'infer',0,0)
+    @extract_visitor = Visitor.new(nil, 'extract',0,0)
   end
 
   # Answers 'can an instance of type t2 be assigned to a variable of type t'.
@@ -205,11 +179,12 @@ class TypeCalculator
       t = type(t)
     end
     t.is_a?(PAnyType) ? t.assignable?(t2) : false
- end
+  end
 
   # Returns an iterable if the t represents something that can be iterated
   def enumerable(t)
-    Puppet.deprecation_warning('TypeCalculator.enumerable is deprecated. Use iterable')
+    #TRANSLATOR 'TypeCalculator.enumerable' and 'iterable' are methods and should not be translated
+    Puppet.deprecation_warning(_('TypeCalculator.enumerable is deprecated. Use iterable'))
     iterable(t)
   end
 
@@ -258,13 +233,13 @@ class TypeCalculator
     when c == FalseClass, c == TrueClass
       type = PBooleanType::DEFAULT
     when c == Class
-      type = PType::DEFAULT
+      type = PTypeType::DEFAULT
     when c == Array
-      # Assume array of data values
-      type = PArrayType::DATA
+      # Assume array of any
+      type = PArrayType::DEFAULT
     when c == Hash
-      # Assume hash with scalar keys and data values
-      type = PHashType::DATA
+      # Assume hash of any
+      type = PHashType::DEFAULT
    else
       type = PRuntimeType.new(:ruby, c.name)
     end
@@ -283,19 +258,19 @@ class TypeCalculator
   #
   def infer(o)
     # Optimize the most common cases into direct calls.
-    case o
-    when String
+    # Explicit if/elsif/else is faster than case
+    if o.is_a?(String)
       infer_String(o)
-    when Integer
+    elsif o.is_a?(Integer) # need subclasses for Ruby < 2.4
       infer_Integer(o)
-    when Array
+    elsif o.is_a?(Array)
       infer_Array(o)
-    when Hash
+    elsif o.is_a?(Hash)
       infer_Hash(o)
-    when Evaluator::PuppetProc
+    elsif o.is_a?(Evaluator::PuppetProc)
       infer_PuppetProc(o)
     else
-      @@infer_visitor.visit_this_0(self, o)
+      @infer_visitor.visit_this_0(self, o)
     end
   end
 
@@ -307,15 +282,14 @@ class TypeCalculator
   # @api public
   #
   def infer_set(o)
-    case o
-      when Array
-        infer_set_Array(o)
-      when Hash
-        infer_set_Hash(o)
-      when Semantic::Version
-        infer_set_Version(o)
-      else
-        infer_set_Object(o)
+    if o.instance_of?(Array)
+      infer_set_Array(o)
+    elsif o.instance_of?(Hash)
+      infer_set_Hash(o)
+    elsif o.instance_of?(SemanticPuppet::Version)
+      infer_set_Version(o)
+    else
+      infer(o)
     end
   end
 
@@ -389,13 +363,13 @@ class TypeCalculator
     # when both are hashes, return a hash with common key- and element type
     if t1.is_a?(PHashType) && t2.is_a?(PHashType)
       key_type = common_type(t1.key_type, t2.key_type)
-      element_type = common_type(t1.element_type, t2.element_type)
-      return PHashType.new(key_type, element_type)
+      value_type = common_type(t1.value_type, t2.value_type)
+      return PHashType.new(key_type, value_type)
     end
 
     # when both are host-classes, reduce to PHostClass[] (since one was not assignable to the other)
-    if t1.is_a?(PHostClassType) && t2.is_a?(PHostClassType)
-      return PHostClassType::DEFAULT
+    if t1.is_a?(PClassType) && t2.is_a?(PClassType)
+      return PClassType::DEFAULT
     end
 
     # when both are resources, reduce to Resource[T] or Resource[] (since one was not assignable to the other)
@@ -414,19 +388,29 @@ class TypeCalculator
       return PFloatType.new([t1.numeric_from, t2.numeric_from].min, [t1.numeric_to, t2.numeric_to].max)
     end
 
-    if t1.is_a?(PStringType) && t2.is_a?(PStringType)
-      common_size_type = common_type(t1.size_type, t2.size_type) unless t1.size_type.nil? || t2.size_type.nil?
-      common_strings = t1.values.empty? || t2.values.empty? ? [] : t1.values | t2.values
-      return PStringType.new(common_size_type, common_strings)
+    if t1.is_a?(PStringType) && (t2.is_a?(PStringType) || t2.is_a?(PEnumType))
+      if(t2.is_a?(PEnumType))
+        return t1.value.nil? ? PEnumType::DEFAULT : PEnumType.new(t2.values | [t1.value])
+      end
+
+      if t1.size_type.nil? || t2.size_type.nil?
+        return t1.value.nil? || t2.value.nil? ? PStringType::DEFAULT : PEnumType.new([t1.value, t2.value])
+      end
+
+      return PStringType.new(common_type(t1.size_type, t2.size_type))
     end
 
     if t1.is_a?(PPatternType) && t2.is_a?(PPatternType)
       return PPatternType.new(t1.patterns | t2.patterns)
     end
 
-    if t1.is_a?(PEnumType) && t2.is_a?(PEnumType)
+    if t1.is_a?(PEnumType) && (t2.is_a?(PStringType) || t2.is_a?(PEnumType))
       # The common type is one that complies with either set
-      return PEnumType.new(t1.values | t2.values)
+      if t2.is_a?(PEnumType)
+        return PEnumType.new(t1.values | t2.values)
+      end
+
+      return t2.value.nil? ? PEnumType::DEFAULT : PEnumType.new(t1.values | [t2.value])
     end
 
     if t1.is_a?(PVariantType) && t2.is_a?(PVariantType)
@@ -437,7 +421,7 @@ class TypeCalculator
     if t1.is_a?(PRegexpType) && t2.is_a?(PRegexpType)
       # if they were identical, the general rule would return a parameterized regexp
       # since they were not, the result is a generic regexp type
-      return PPatternType::DEFAULT
+      return PRegexpType::DEFAULT
     end
 
     if t1.is_a?(PCallableType) && t2.is_a?(PCallableType)
@@ -451,17 +435,25 @@ class TypeCalculator
       return PNumericType::DEFAULT
     end
 
+    if common_scalar_data?(t1, t2)
+      return PScalarDataType::DEFAULT
+    end
+
     if common_scalar?(t1, t2)
       return PScalarType::DEFAULT
     end
 
     if common_data?(t1,t2)
-      return PDataType::DEFAULT
+      return TypeFactory.data
     end
 
     # Meta types Type[Integer] + Type[String] => Type[Data]
-    if t1.is_a?(PType) && t2.is_a?(PType)
-      return PType.new(common_type(t1.type, t2.type))
+    if t1.is_a?(PTypeType) && t2.is_a?(PTypeType)
+      return PTypeType.new(common_type(t1.type, t2.type))
+    end
+
+    if common_rich_data?(t1,t2)
+      return TypeFactory.rich_data
     end
 
     # If both are Runtime types
@@ -492,30 +484,13 @@ class TypeCalculator
   # Produces the superclasses of the given class, including the class
   def superclasses(c)
     result = [c]
-    while s = c.superclass
+    while s = c.superclass #rubocop:disable Lint/AssignmentInCondition
       result << s
       c = s
     end
     result
   end
 
-  # Produces a string representing the type
-  # @api public
-  #
-  def string(t)
-    Puppet.deprecation_warning('TypeCalculator.string is deprecated. Use TypeFormatter')
-    TypeFormatter.singleton.string(t)
-  end
-
-  # Produces a debug string representing the type (possibly with more information that the regular string format)
-  # @api public
-  #
-  def debug_string(t)
-    Puppet.deprecation_warning('TypeCalculator.debug_string is deprecated. Use TypeFormatter')
-    TypeFormatter.singleton.debug_string(t)
-  end
-
-
   # Reduces an enumerable of types to a single common type.
   # @api public
   #
@@ -530,11 +505,11 @@ class TypeCalculator
     reduce_type(enumerable.map {|o| infer(o) })
   end
 
-  # The type of all modules is PType
+  # The type of all modules is PTypeType
   # @api private
   #
   def infer_Module(o)
-    PType::new(PRuntimeType.new(:ruby, o.name))
+    PTypeType::new(PRuntimeType.new(:ruby, o.name))
   end
 
   # @api private
@@ -555,33 +530,38 @@ class TypeCalculator
   # @api private
   def infer_Object(o)
     if o.is_a?(PuppetObject)
-      o._ptype
+      o._pcore_type
     else
       name = o.class.name
+      return PRuntimeType.new(:ruby, nil) if name.nil? # anonymous class that doesn't implement PuppetObject is impossible to infer
       ir = Loaders.implementation_registry
       type = ir.nil? ? nil : ir.type_for_module(name)
-      type.nil? ? PRuntimeType.new(:ruby, name) : type
+      return PRuntimeType.new(:ruby, name) if type.nil?
+      if type.is_a?(PObjectType) && type.parameterized?
+        type = PObjectTypeExtension.create_from_instance(type, o)
+      end
+      type
     end
   end
 
-  # The type of all types is PType
+  # The type of all types is PTypeType
   # @api private
   #
   def infer_PAnyType(o)
-    PType.new(o)
+    PTypeType.new(o)
   end
 
-  # The type of all types is PType
+  # The type of all types is PTypeType
   # This is the metatype short circuit.
   # @api private
   #
-  def infer_PType(o)
-    PType.new(o)
+  def infer_PTypeType(o)
+    PTypeType.new(o)
   end
 
   # @api private
   def infer_String(o)
-    PStringType.new(size_as_type(o), [o])
+    PStringType.new(o)
   end
 
   # @api private
@@ -596,7 +576,7 @@ class TypeCalculator
 
   # @api private
   def infer_Regexp(o)
-    PRegexpType.new(o.source)
+    PRegexpType.new(o)
   end
 
   # @api private
@@ -620,9 +600,8 @@ class TypeCalculator
       max += 1
       PAnyType::DEFAULT
     end
-    mapped_types << min
-    mapped_types << max
-    TypeFactory.callable(*mapped_types)
+    param_types = Types::PTupleType.new(mapped_types, Types::PIntegerType.new(min, max))
+    Types::PCallableType.new(param_types)
   end
 
   # @api private
@@ -643,14 +622,34 @@ class TypeCalculator
     end
   end
 
+  # @api private
+  def infer_Sensitive(o)
+    PSensitiveType.new(infer(o.unwrap))
+  end
+
+  # @api private
+  def infer_Timespan(o)
+    PTimespanType.new(o, o)
+  end
+
+  # @api private
+  def infer_Timestamp(o)
+    PTimestampType.new(o, o)
+  end
+
   # @api private
   def infer_TrueClass(o)
-    PBooleanType::DEFAULT
+    PBooleanType::TRUE
   end
 
   # @api private
   def infer_FalseClass(o)
-    PBooleanType::DEFAULT
+    PBooleanType::FALSE
+  end
+
+  # @api private
+  def infer_URI(o)
+    PURIType.new(o)
   end
 
   # @api private
@@ -661,18 +660,27 @@ class TypeCalculator
     # A mapping must be made to empty string. A nil value will result in an error later
     title = o.title
     title = '' if :undef == title
-    PType.new(PResourceType.new(o.type.to_s, title))
+    PTypeType.new(PResourceType.new(o.type.to_s, title))
   end
 
   # @api private
   def infer_Array(o)
-    if o.empty?
-      PArrayType::EMPTY
+    if o.instance_of?(Array)
+      if o.empty?
+        PArrayType::EMPTY
+      else
+        PArrayType.new(infer_and_reduce_type(o), size_as_type(o))
+      end
     else
-      PArrayType.new(infer_and_reduce_type(o), size_as_type(o))
+      infer_Object(o)
     end
   end
 
+  # @api private
+  def infer_Binary(o)
+    PBinaryType::DEFAULT
+  end
+
   # @api private
   def infer_Version(o)
     PSemVerType::DEFAULT
@@ -685,12 +693,16 @@ class TypeCalculator
 
   # @api private
   def infer_Hash(o)
-    if o.empty?
-      PHashType::EMPTY
+    if o.instance_of?(Hash)
+      if o.empty?
+        PHashType::EMPTY
+      else
+        ktype = infer_and_reduce_type(o.keys)
+        etype = infer_and_reduce_type(o.values)
+        PHashType.new(ktype, etype, size_as_type(o))
+      end
     else
-      ktype = infer_and_reduce_type(o.keys)
-      etype = infer_and_reduce_type(o.values)
-      PHashType.new(ktype, etype, size_as_type(o))
+      infer_Object(o)
     end
   end
 
@@ -716,7 +728,7 @@ class TypeCalculator
     if o.empty?
       PHashType::EMPTY
     elsif o.keys.all? {|k| PStringType::NON_EMPTY.instance?(k) }
-      PStructType.new(o.each_pair.map { |k,v| PStructElement.new(PStringType.new(size_as_type(k), [k]), infer_set(v)) })
+      PStructType.new(o.each_pair.map { |k,v| PStructElement.new(PStringType.new(k), infer_set(v)) })
     else
       ktype = PVariantType.maybe_create(o.keys.map {|k| infer_set(k) })
       etype = PVariantType.maybe_create(o.values.map {|e| infer_set(e) })
@@ -726,7 +738,7 @@ class TypeCalculator
 
   # @api private
   def infer_set_Version(o)
-    PSemVerType.new(Semantic::VersionRange.new(o, o))
+    PSemVerType.new([SemanticPuppet::VersionRange.new(o, o)])
   end
 
   def unwrap_single_variant(possible_variant)
@@ -790,8 +802,18 @@ class TypeCalculator
 
   private
 
+  def common_rich_data?(t1, t2)
+    d = TypeFactory.rich_data
+    d.assignable?(t1) && d.assignable?(t2)
+  end
+
   def common_data?(t1, t2)
-    PDataType::DEFAULT.assignable?(t1) && PDataType::DEFAULT.assignable?(t2)
+    d = TypeFactory.data
+    d.assignable?(t1) && d.assignable?(t2)
+  end
+
+  def common_scalar_data?(t1, t2)
+    PScalarDataType::DEFAULT.assignable?(t1) && PScalarDataType::DEFAULT.assignable?(t2)
   end
 
   def common_scalar?(t1, t2)
