@@ -6,8 +6,8 @@ Puppet::Type.type(:package).provide :openbsd, :parent => Puppet::Provider::Packa
 
     This provider supports the `install_options` and `uninstall_options`
     attributes, which allow command-line flags to be passed to pkg_add and pkg_delete.
-    These options should be specified as a string (e.g. '--flag'), a hash (e.g. {'--flag' => 'value'}),
-    or an array where each element is either a string or a hash."
+    These options should be specified as an array where each element is either a
+     string or a hash."
 
   commands :pkginfo   => "pkg_info",
            :pkgadd    => "pkg_add",
@@ -25,49 +25,47 @@ Puppet::Type.type(:package).provide :openbsd, :parent => Puppet::Provider::Packa
   mk_resource_methods
 
   def self.instances
-    # our regex for matching pkg_info output
-    regex = /^(.*)-(\d[^-]*)[-]?(\w*)(.*)$/
     packages = []
 
     begin
-      packages = pkginfo('-a')
-      packages.split("\n").collect do |package|
-        if match = regex.match(package.split[0])
-          new( :name => match.captures[0],
-               :ensure => match.captures[1],
-               :flavor => match.captures[2],
-          )
-        else
-          unless line =~ /Updating the pkgdb/
-            # Print a warning on lines we can't match, but move
-            # on, since it should be non-fatal
-            warning("Failed to match line #{line}")
+      execpipe(listcmd) do |process|
+        # our regex for matching pkg_info output
+        regex = /^(.*)-(\d[^-]*)[-]?([\w-]*)(.*)$/
+        fields = [:name, :ensure, :flavor ]
+        hash = {}
+
+        # now turn each returned line into a package object
+        process.each_line { |line|
+          match = regex.match(line.split[0])
+          if match
+            fields.zip(match.captures) { |field,value|
+              hash[field] = value
+            }
+
+            hash[:provider] = self.name
+
+            packages << new(hash)
+            hash = {}
           end
-        end
+        }
       end
+
+      return packages
     rescue Puppet::ExecutionFailure
       return nil
     end
   end
 
-  def self.prefetch(resources)
-    packages = instances
-    resources.keys.each do |name|
-      if provider = packages.find{ |pkg| pkg.name == name }
-        resources[name].provider = provider
-      end
-    end
-  end
-
-  def flavor=(value)
-    install
+  def self.listcmd
+    [command(:pkginfo), "-a"]
   end
 
   def latest
+
     if @resource[:flavor]
       query = "#{@resource[:name]}--#{@resource[:flavor]}"
     else
-      query = @resource[:name]
+      query = @resource[:name] + "--"
     end
 
     output = Puppet::Util.withenv({}) {pkginfo "-Q", query}
@@ -136,11 +134,12 @@ Puppet::Type.type(:package).provide :openbsd, :parent => Puppet::Provider::Packa
     # installing with 'latest', we do need to handle the flavors. This is
     # done so we can feed pkg_add(8) the full package name to install to
     # prevent ambiguity.
-    if latest && resource[:flavor]
+    if resource[:flavor]
       "#{resource[:name]}--#{resource[:flavor]}"
-    elsif latest
-      # Don't depend on get_version for updates.
-      @resource[:name]
+    elsif resource[:name].to_s.match(/[a-z0-9]%[0-9a-z]/i)
+        "#{resource[:name]}"
+    elsif not latest
+      "#{resource[:name]}--"
     else
       # If :ensure contains a version, use that instead of looking it up.
       # This allows for installing packages with the same stem, but multiple
@@ -148,35 +147,45 @@ Puppet::Type.type(:package).provide :openbsd, :parent => Puppet::Provider::Packa
       if @resource[:ensure].to_s =~ /(\d[^-]*)$/
         use_version = @resource[:ensure]
       else
- 
         use_version = get_version
       end
 
-      [ @resource[:name], use_version, @resource[:flavor]].join('-').gsub(/-+$/, '')
+      if resource[:flavor]
+        [ @resource[:name], use_version, @resource[:flavor]].join('-').gsub(/-+$/, '')
+      else
+        [ @resource[:name], use_version ]
+      end
     end
   end
 
   def get_version
+    pkg_search_name = @resource[:name]
+    unless pkg_search_name.match(/[a-z0-9]%[0-9a-z]/i) and not @resource[:flavor]
+      # we are only called when no flavor is specified
+      # so append '--' to the :name to avoid patch versions on flavors
+      pkg_search_name << "--"
+    end
     # our regex for matching pkg_info output
     regex = /^(.*)-(\d[^-]*)[-]?(\w*)(.*)$/
     master_version = 0
     version = -1
 
     # pkg_info -I might return multiple lines, i.e. flavors
-    matching_pkgs = pkginfo("-I", @resource[:name])
+    matching_pkgs = pkginfo("-I", "pkg_search_name")
     matching_pkgs.each_line do |line|
       if match = regex.match(line.split[0])
         # now we return the first version, unless ensure is latest
         version = match.captures[1]
         return version unless @resource[:ensure] == "latest"
-
         master_version = version unless master_version > version
       end
+
     end
 
     return master_version unless master_version == 0
     return '' if version == -1
     raise Puppet::Error, _("%{version} is not available for this package") % { version: version }
+
   rescue Puppet::ExecutionFailure
     return nil
   end
