@@ -35,7 +35,8 @@ class Puppet::Transaction::ResourceHarness
 
   def scheduled?(resource)
     return true if Puppet[:ignoreschedules]
-    return true unless schedule = schedule(resource)
+    schedule = schedule(resource)
+    return true unless schedule
 
     # We use 'checked' here instead of 'synced' because otherwise we'll
     # end up checking most resources most times, because they will generally
@@ -51,7 +52,8 @@ class Puppet::Transaction::ResourceHarness
       return nil
     end
 
-    return nil unless name = resource[:schedule]
+    name = resource[:schedule]
+    return nil unless name
     resource.catalog.resource(:schedule, name) || resource.fail(_("Could not find schedule %{name}") % { name: name })
   end
 
@@ -99,7 +101,7 @@ class Puppet::Transaction::ResourceHarness
   # We persist the last known values for the properties of a resource after resource
   # application.
   # @param [Puppet::Type] resource resource whose values we are to persist.
-  # @param [ResourceApplicationContent] context the application context to operate on.
+  # @param [ResourceApplicationContext] context the application context to operate on.
   def persist_system_values(resource, context)
     param_to_event = {}
     context.status.events.each do |ev|
